from moteur_avec_variables.proposition_avec_variables import *
from .noeud import Noeud

class NoeudsTestes:
    """ Un objet destiné à recueillir les noeuds testés.

       Vérifier si un noeud a déjà été testé est une tâche non triviale, qui\
       met en jeu plusieurs opérations. Par souci de propreté du code, et afin\ 
       d'abstraire le processus, les méthodes nécessaires sont donc regroupées\
       dans cette classe.
    """
    
    echec = 'échec'

    def __init__(self):
        # Nous utiliserons une liste standard pour stocker les descriptions des\
        # noeuds déjà testés.
        self.descriptions = []

    def ajoute(self, noeud):
        """ Enregistre un noeud comme testé. """

        description = noeud.description_standardisee()
        self.descriptions.append(description)

    def __contains__(self, noeud):
        """ Permet d'utiliser la syntaxe ``noeud in noeuds_testes`` pour\
            déterminer si un noeud a déjà été testé.
        """

        if not isinstance(noeud, Noeud):
            raise ValueError("Seul un noeud peut être testé.") 

        description = noeud.description_standardisee()
        for descr in self.descriptions:
            if self.inclut(descr, description):
                return True

        return False
            
    def match(self, prop1, prop2, env):
        """ Tente de trouver un match entre deux propositions étant donné un\
            environnement.
        """

        prop1 = prop1[:]
        prop2 = prop2[:]
        env = env.copy()

        # Si les deux propositions sont vides, le processus est terminé.
        if len(prop1) == 0 and len(prop2) == 0:
            return env

        # Si une des propositions seulement est vide, les deux n'ont pas la même
        # longueur, donc elles sont bien différentes.
        elif len(prop1) == 0 or len(prop2) == 0:
            return NoeudsTestes.echec

        # Si les deux sont des variables,
        elif est_une_variable(prop1) and est_une_variable(prop2):
            # on teste si la substitution a déjà été trouvée
            if prop1 in env:
                if env[prop1] == prop2:
                    return env

            # sinon on l'ajoute a l'environnement courant.
            else:
                if prop1 != prop2:
                    env[prop1] = prop2
                return env
                
        # Si l'une des propositions est un atome, on retourne l'environnement.
        elif est_atomique(prop1) or est_atomique(prop2):
            if prop1 == prop2:
                return env
        else:
            tete1 = tete(prop1)
            reste1 = corps(prop1)
            tete2 = tete(prop2)
            reste2 = corps(prop2)
            env_tete = self.match(tete1, tete2, env)
            if env_tete == NoeudsTestes.echec:
                return NoeudsTestes.echec
            env_reste = self.match(reste1, reste2, env_tete)
            if env_reste == NoeudsTestes.echec:
                return NoeudsTestes.echec
            return env_reste

        # Aucune tentative n'a donné de résultat.
        return NoeudsTestes.echec

    def inclut_sous_buts(self, sous_buts1, sous_buts2, env):
        """ Tente de trouver un match entre des sous-buts étant donné un\
            environnement. 
        """

        if len(sous_buts1) == 0:
            return True
        
        sb1 = tete(sous_buts1)
        sous_buts_restants1 = corps(sous_buts1)
    
        for sb2 in sous_buts2:
            nouvel_env = self.match(sb1, sb2, env)
            if nouvel_env != NoeudsTestes.echec:
                # On teste le match avec les sous-buts restants.
                sous_buts_restants2 = [sb for sb in sous_buts2 if sb != sb2]
                if self.inclut_sous_buts(sous_buts_restants1, sous_buts_restants2, nouvel_env):
                    return True

        return False

    def inclut(self, descr1, descr2):
        """ Tente de trouver un match entre deux descriptions afin de\
            prouver leur équivalence.
        """

        env = self.match(tete(descr1), tete(descr2), {})
        if env != NoeudsTestes.echec:
            # On compare les sous-buts non-résolus.
            return self.inclut_sous_buts(corps(descr1), corps(descr2), env)
        return False